/* 
   Security loading modules support 

   Copyright 1999, 2000 by Xie Huagang ( xhg@gem.ncic.ac.cn).

   This file is part of the Linux Intrusion Detection System.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   complied with 

	gcc -O6 -c check_modules.c -fomit-frame-pointer  -Wall -I/home/software/kernel/2.2.12/linux/include -DMODULE -D__KERNEL__

  */

#include <linux/slab.h>
#include <linux/file.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/malloc.h>

#include <linux/kernel.h>
#include <asm/unistd.h>
#include <sys/syscall.h>
#include <asm/uaccess.h>
#include <asm/segment.h>
#include <linux/mm.h>

#include <linux/utsname.h>

typedef int (*sysfun_p)();
extern sysfun_p sys_call_table[];

extern int ids_load;
extern struct new_utsname system_utsname;

char *version;
/*must be defined because of syscall macro used below*/
int errno;

int	check_started =0;
extern int ids_load;
extern int lids_started;
/*we define our own systemcall*/
int __NR_myexecve;

static inline _syscall1(int, brk, void *, end_data_segment);

int (*orig_execve) (const char *, const char *[], const char *[]);

int stat_file(const char *filename)
{
	struct file *filp;

        filp = filp_open(filename,O_RDONLY,O_RDONLY);
        if (IS_ERR(filp)||(filp==NULL)) {
		return -1;
        }

	return 0;
}

static char *
search_module_path(const char *base)
{
  static const char default_path[] =
    ".:"
    "/linux/modules:"
    "/lib/modules/%s/fs:"
    "/lib/modules/%s/net:"
    "/lib/modules/%s/scsi:"
    "/lib/modules/%s/block:"
    "/lib/modules/%s/cdrom:"
    "/lib/modules/%s/ipv4:"
    "/lib/modules/%s/ipv6:"
    "/lib/modules/%s/sound:"
    "/lib/modules/%s/fc4:"
    "/lib/modules/%s/video:"
    "/lib/modules/%s/misc:"
    "/lib/modules/default/fs:"
    "/lib/modules/default/net:"
    "/lib/modules/default/scsi:"
    "/lib/modules/default/block:"
    "/lib/modules/default/cdrom:"
    "/lib/modules/default/ipv4:"
    "/lib/modules/default/ipv6:"
    "/lib/modules/default/sound:"
    "/lib/modules/default/fc4:"
    "/lib/modules/default/video:"
    "/lib/modules/default/misc:"
    "/lib/modules/fs:"
    "/lib/modules/net:"
    "/lib/modules/scsi:"
    "/lib/modules/block:"
    "/lib/modules/cdrom:"
    "/lib/modules/ipv4:"
    "/lib/modules/ipv6:"
    "/lib/modules/sound:"
    "/lib/modules/fc4:"
    "/lib/modules/video:"
    "/lib/modules/misc";

  char *path, *p, *filename;
  size_t len;

	
   if (stat_file(base) == 0 ){
  	filename = kmalloc(strlen(base)+1, GFP_KERNEL);
	memset(filename,'\0',strlen(base)+1);
	strcpy(filename,base);
	return filename;
   }
/*
  if ((path = getenv("MODPATH")) == NULL)
*/
    path = (char *)default_path;

  /* Make a copy so's we can mung it with strtok.  */
  len = strlen(path)+1;
 
  p = kmalloc(len, GFP_KERNEL);
  path = memcpy(p, path, len);

  filename = kmalloc(PATH_MAX, GFP_KERNEL);
  memset(filename,'\0',PATH_MAX);

  for (p = strtok(path, ":"); p != NULL ; p = strtok(NULL, ":"))
    {

      len = sprintf(filename, p, version);
      len += sprintf(filename+len,  "/%s", base);

      if ( stat_file(filename) ==0 )
	return filename;

      sprintf(filename+len,  ".o");

      if (stat_file(filename) == 0 )
	return filename;
    }

  kfree(filename);
  return NULL;
}
/*
 *	check_insmod_argv.
 *
 */
int  check_insmod_argv( const char **argv , const char **envp )
{
	int i=1,argc=1;
	char *file;

	while(argv[argc] != NULL)
		argc++;

	for(i=1;i<argc;i++) {
	
		if(!strncmp(argv[i],"--name",6) || !strncmp(argv[i],"-o",2 )){
			i++; 
			continue;
		}
		if(argv[i][0] == '-' ) continue;
			
		if ( (file=search_module_path( argv[i])) != NULL ) {
			if(!strncmp(file,"/lib/modules/",
					strlen("/lib/modules/"))) {
				return 0;
			}
			break;
		} 
	}
	return -1;
}


char *strncpy_fromfs(char *dest, const char *src, int n)
{
 int compt = 0;
 int i=0;
 char k;

 do {
  __get_user(k,src+i);
  dest[compt]=k;
  compt++;
  i++;
 } while ((dest[compt - 1] != '\0') && (compt != n));
 return dest;
}


/*this is something like a systemcall macro called with SYS_execve, the
asm code calls int 0x80 with the registers set in a way needed for our own
__NR_myexecve systemcall*/

int my_execve(const char *filename, const char *argv[], const char *envp[])
{
    long __res;

    __asm__ volatile ("int $0x80":"=a" (__res):"0"(__NR_myexecve), "b"((long) 
                     (filename)), "c"((long) (argv)),"d"((long) (envp)));
    return (int) __res;
}


int hacked_execve(const char *filename, const char *argv[], const char *envp[])
{
 char *test;
 int ret;
 char *truc = "/sbin/insmod";        /*the file we *should* be executed*/

 if(ids_load) {
 	test = (char *) kmalloc(strlen(truc) + 2, GFP_KERNEL);

 	(void) strncpy_fromfs(test, filename, strlen(truc));
 	test[strlen(truc)] = '\0';

 	if (!strcmp(test, truc)) {
		if ( check_insmod_argv( argv , envp ) < 0 ) { 
 			kfree(test);
			return -EACCES ;
		}
 	}
 	if (!strcmp(test, "/sbin/reboot")&& ids_load) {
 		kfree(test);
		return -EACCES;
 	}
 	kfree(test);
 }
 ret = my_execve(filename, argv, envp);
 return ret;
}

int init_module(void)                /*module setup*/
{
 	__NR_myexecve = 200;

	if (!lids_started)
		return -EPERM;

 	version = system_utsname.release;

 	orig_execve = sys_call_table[SYS_execve];

 	if (__NR_myexecve != 0) 
 	{
  		sys_call_table[__NR_myexecve] = orig_execve; 
  		sys_call_table[SYS_execve] = (void *) hacked_execve;
 	}
	check_started = 1;
 	return 0;
}

void cleanup_module(void)            /*module shutdown*/
{
 	sys_call_table[SYS_execve]=orig_execve;
}
