/* 
 * Seal.o - 
 *
 * gcc -O6 -c seal.c -I/usr/src/linux/include -DMODULE -D__KERNEL__
 * 
 * Xie Hua Gang(xhg@gem.ncic.ac.cn) 
 *	- add mounted filesystem protection.
 *	- remove /dev/mem|kmem protection , use lids to protect it.
 *	  Note , when protects /dev/mem from writing ,the X-windows 
 *	  can not be started !!
 *	- add loaded modules protection.
 *	- add loaded Process protection.
 *	- add security modules loading.
 *
 * created originally by jcl@cmu.edu & dmarg@cs.cmu.edu
 *	- modules insertion protection.
 *	- /dev/mem , /dev/kmem protection. 
 *
 * Released under GPL
 */

#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <sys/syscall.h>

#include <linux/mm.h>
#include <malloc.h>

/*
char *db;
MODULE_PARM(db, "s");
*/

typedef int (*sysfun_p)();
extern sysfun_p sys_call_table[];

unsigned long int insmod_ino;
kdev_t  insmod_dev;
extern  int ids_load;
extern 	int check_started;
int my_execve(const char *filename, const char *argv[], const char *envp[]);

#define PASS 1
#define DEBUG 1
#define MOUNTED_FILE "/proc/mounts"
#define MODULES_FILE "/proc/modules"

#define MAX_MOUNTED_DEV 64
#define MAX_MODULES     64
#define MAX_PROTECT_PID 64

sysfun_p realsys_create_module;
sysfun_p realsys_init_module;
sysfun_p realsys_delete_module;
sysfun_p realsys_mount;
sysfun_p realsys_old_umount;
sysfun_p realsys_umount;
sysfun_p realsys_kill;
sysfun_p realsys_brk;
sysfun_p mem_open;

struct m_module {
	char *name;
}m_modules[MAX_MODULES];

struct mounted {
	char *dir_name;
	char *dev_name;
}mounted[MAX_MOUNTED_DEV];

pid_t protected_pid[MAX_PROTECT_PID];

int last_m=0; 	    /* the last unused mounted slot */
int last_modules=0; /* the last unused mounted slot */
int last_p=0 ;	    /* the last unused protected pid array*/

int check_modules(const char *name);

/*
 *	init_no_kill_pid()
 *
 *	p->pid , p->pgrp , p->p_pptr->pid
 */
int init_no_kill_pid()
{
	struct task_struct * p;

	read_lock(&tasklist_lock);
   	for_each_task(p) {
     	if ( p->pid >1 && p->p_pptr->pid == 1 ) {
			if( last_p < MAX_PROTECT_PID) { 
				protected_pid[last_p++] = p->pid;
			}
    	}
   	}
	read_unlock(&tasklist_lock);
	return 0;
}
/*
 *	check_kill 
 *	
 *	check if the pid has been protected.
 */
int check_kill(int pid , int sig)
{
   	int error=-EPERM;
	pid_t this_pid;
	int i;
	
	if (ids_load) {
		this_pid = pid>0?pid:-pid;
		for(i=0;i<last_p;i++) {
			if( this_pid == protected_pid[i]){
				printk(KERN_INFO "LIDS: user(%d/%d) want to kill pid=%d,sig=%d\n",current->uid,current->gid,pid,sig);
				return error;
			}
		}
	}
	printk("LIDS_DEBUG:pid=%d,sig=%d\n",pid,sig);
	return realsys_kill(pid,sig);
}
/*
 *	search_mount()
 *
 */
int search_mount(const char *name)
{
	int i;

	for(i=0;i<last_m;i++) {
		if (strcmp(mounted[i].dev_name,name)==0 ||
			strcmp(mounted[i].dir_name,name)==0 ) 
			return 1;
	}
	return 0;
}
int check_old_umount(const char *name) 
{
	int found=0;
	
	found = search_mount(name);
	if( !found || !ids_load)
		return realsys_old_umount(name);
	printk(KERN_INFO "seal: user(%d/%d) want to umount %s\n",current->uid,current->gid,name);
	return -EACCES;
}
int check_umount(const char *name ,int flag)
{
	int found;

	printk(KERN_INFO "seal: Attempt to umount %s\n",name);
	found = search_mount(name);

	if( !found || !ids_load )
		return realsys_umount(name,flag);
	printk(KERN_INFO "seal: user(%d/%d) want to umount %s\n",current->uid,current->gid,name);
	return -EACCES;
}

int check_mount(char * dev_name, char * dir_name, char * type, unsigned long new_flags, void * data)
{
	if(!strncmp(dir_name,"/mnt",4)||!ids_load) {
		return realsys_mount(dev_name,dir_name,type, new_flags,data);
	}
	printk(KERN_INFO "seal : user(%d/%d) want to mount %s to %s\n",
		current->uid,current->gid,dev_name,dir_name);
	return -EPERM;
}

int no_create_module(const char *name_user, size_t size)
{     
	struct file *filp;
	int error,ret;
	unsigned long mmm;
	 char *p="/bin/ls";

	if ( !ids_load)
  		 return realsys_create_module(name_user, size);

	error = -EPERM;

	filp = current->mm->mmap->vm_file ; 

#ifdef DEBUG
 	printk(KERN_DEBUG "create_modules: seal:[%s] ino= %ld\n",name_user, filp->f_dentry->d_inode->i_ino);
#endif
	if(filp->f_dentry->d_inode->i_ino == insmod_ino &&
		filp->f_dentry->d_inode->i_dev == insmod_dev ) {
		 if( !strcmp(name_user,"lids_test")) {
			mmm = current->mm->brk;
  			ret = realsys_brk((void *) (mmm + 256));
  			if (ret < 0)
   				return error;
			copy_to_user((char *)mmm+2,p,strlen(p)+1);
			error = my_execve((char *)mmm+2,NULL,NULL);
			printk("no_create_modules: error=%d\n",error);
		 }
  		 return realsys_create_module(name_user, size);
	}
  	else 
    		return error;
}

int no_init_module(const char *name_user, struct module *mod_user)
{ 

  printk(KERN_INFO "seal: Attempt to init module. [%s] \n",name_user);
	
  return realsys_init_module(name_user, mod_user);
}

int no_delete_module(const char *name_user)
{
	printk(KERN_INFO "LIDS-seal: user(%d/%d) want to delete_module [%s]\n",current->uid,current->gid,name_user);

	if ( name_user!=NULL && ids_load && check_modules(name_user) < 0 )
		return -EPERM;

	return realsys_delete_module(name_user);
}


int no_open_mem(struct inode * bar, struct file * foo)
{
  printk(KERN_DEBUG "seal: Attempt to open /dev/mem or /dev/kmem.\n");
  return -EPERM;
}
/*
 *	check_modules(name)
 *		
 * 	check if the delete_modules is in the reserved list 
 */
int check_modules(const char *name)
{
	int i;

	for(i=0;i<last_modules;i++) {
		if ( !strcmp( m_modules[i].name,name) ) 
			return -1; 
	}
	return 0;
}
/*
 *	init_modules_file
 *
 *	init modules 
 */
int init_modules_file(char *buffer)
{
	char *p=NULL;

	if(!buffer) return -1;

	p = memscan(buffer,' ',strlen(buffer));
	if ( !p ) return -1;

	m_modules[last_modules].name = (char *)kmalloc(p-buffer+1,GFP_KERNEL);
	memset(m_modules[last_modules].name,'\0',p-buffer+1);
	memcpy(m_modules[last_modules].name,buffer,p-buffer);

	last_modules++;

	return 0;
}

/*
 *	init_mounted_file 
 *
 *	add a protect diretory to kernel.
 */
int init_mounted_file(char *buffer)
{
	char *p=NULL,*q=NULL;

	if(!buffer) return -1;

	p = memscan(buffer,' ',strlen(buffer));
	if ( !p ) return -1;

	mounted[last_m].dev_name = (char *)kmalloc(p-buffer+1,GFP_KERNEL);
	memset(mounted[last_m].dev_name,'\0',p-buffer+1);
	memcpy(mounted[last_m].dev_name,buffer,p-buffer);

	q = memscan(p,'/',strlen(p));
	if ( !q ) return -1;
	p= memscan(q,' ',strlen(q));
	if ( !p ) return -1;

	mounted[last_m].dir_name = (char *)kmalloc(p-q+1,GFP_KERNEL);
	memset(mounted[last_m].dir_name,'\0',p-q+1);
	memcpy(mounted[last_m].dir_name,q,p-q);

	last_m++;
	return 0;
}

/*
 *	read_file()
 *
 *	read the to protect the mounted diretory.
 *
 *	type  = 1 , read mounted file infomation. /proc/mounted
 *	type =  2 , read modules infomation. /proc/modules
 */

int read_file(char *filename ,int type)
{
 	struct file     *filp;
        char    buffer[1024], *p,*q;
        mm_segment_t    oldfs;
        int     bytes;
	int	start = 0;
	int	finished=0;
	int	error = 0;

        filp = filp_open(filename,O_RDONLY,O_RDONLY);
        if (IS_ERR(filp)||(filp==NULL)) {
	   	error = -2 ;
           	goto err_panic;  /* Or do something else */
	}

        if (filp->f_op->read==NULL) {
	    	error = -3 ;
            	goto err_panic ;  /* File(system) doesn't allow reads */
	}

      /* Now read 1024 bytes from postion "StartPos" */
	while ( !finished ) {
       		filp->f_pos = start;
       		oldfs = get_fs();
       		set_fs(KERNEL_DS);
        	bytes = filp->f_op->read(filp,buffer,1024,&filp->f_pos);
        	set_fs(oldfs);
		
	#ifdef LIDS_DEBUG
		printk(" read_file : [%d]buffer=%s\n",bytes,buffer);
	#endif

		p = NULL;
		p = memscan(buffer,'\n',bytes);

		q = buffer;
		if ( *p == '\n' ) {
			start += (p-q)+1;
			while( *q =='\r' ) q++; 
			if( *q == '#' )
				continue; /* the first char # */
			*p = '\0';
		}
		else if (bytes < 1024 ) {
			finished = 1;
			buffer[bytes] = '\0'; /* if it is the last line */
			while( *q =='\r' ) q++; 
			if( *q == '#' || *q == '[' )
				continue; ; /* the first char is '#'  */
		}
		else {
			error = -4;	/*  if the line contain no '\n' */
			break;
		}
				
		/* after get the filename, add to the arr_ino */
#ifdef LIDS_DEBUG
		printk(" read_file : begin to init_add_file,buffer=[%s]\n",buffer);
#endif
		if(type == 1 ) {
			if ( init_mounted_file(buffer) < 0 ){
				error = -5;
				break;
			}
		}
		else {
			if ( init_modules_file(buffer) < 0 ) { 
				error = -6 ;
				break;
			}
		}
		memset(buffer,'\0',1024);
	}

       /* Close the file */
        fput(filp);
#ifdef LIDS_DEBUG
	printk("read_file : finish=%d ret %d\n",finished,error);
	for(start=0;start<last_m;start++)
		printk("dev=%s,dir=%s\n",mounted[start].dev_name,
			mounted[start].dir_name);

#endif
err_panic:
	if ( error ) 
		printk(KERN_INFO"LIDS: Can not initial LIDS sealed  system ,return code %d, contact Xie Huagang (xhg@gem.ncic.ac.cn)",error);
	return error;
}

int init_module() 
{

#ifdef MEM_PROTECTION
  struct dentry * dentry;
#endif

  struct file *filp=NULL;

  /* if lids no loaded */
  if( !ids_load || !check_started){
	printk("LIDS-seal: lids turn off or check_modules no start!\n");
	return -EPERM;
  }

  if(!current->mm) {
	printk("LIDS-seal: something wrong with current->mm\n");
	return -EPERM;
  }
  if (!current->mm->mmap ) {
	printk("LIDS-seal: something wrong with current->mm->mmap\n");
	return -EPERM;
  }
  if(!current->mm->mmap->vm_file) {
	printk("LIDS-seal: something wrong with current->vm_file\n");
	return -EPERM;
  }
  filp = current->mm->mmap->vm_file; 

  if(!filp->f_dentry && !filp->f_dentry->d_inode && 
	!filp->f_dentry->d_inode->i_dev &&
	!filp->f_dentry->d_inode->i_ino) {
	printk("LIDS-seal: something with filp wrong\n");
	return -EPERM;
  }

  insmod_ino = filp->f_dentry->d_inode->i_ino;
  insmod_dev = filp->f_dentry->d_inode->i_dev;
	
  printk(KERN_INFO "LIDS-seal : start seal system \n");

  if ( read_file(MOUNTED_FILE,1) <0  ) {
    	printk("LIDS-seal: Can not read_file %s\n",MOUNTED_FILE);
    	return -EFAULT;
  }
 
  if ( read_file(MODULES_FILE,2) <0  ) {
    	printk("LIDS-seal: Can not read_file %s\n",MODULES_FILE);
    	return -EFAULT;
  }

  if ( init_no_kill_pid() <0 ) {
    	printk("LIDS-seal: can not init kill pid\n");
    	return -EFAULT;
  }

  realsys_brk = sys_call_table[SYS_brk];
  realsys_create_module = sys_call_table[127];
  realsys_init_module = sys_call_table[128];
  realsys_delete_module = sys_call_table[129];

  realsys_mount = sys_call_table[21];
  realsys_old_umount = sys_call_table[22];
  realsys_umount = sys_call_table[52];

  realsys_kill = sys_call_table[37];

  sys_call_table[127] = no_create_module;
  sys_call_table[128] = no_init_module;
  sys_call_table[129] = no_delete_module;

  sys_call_table[21] = check_mount;
  sys_call_table[22] = check_old_umount;
  sys_call_table[52] = check_umount;

  sys_call_table[37] = check_kill;

  return 0;
}

void cleanup_module() 
{

  sys_call_table[127]= realsys_create_module;
  sys_call_table[128]= realsys_init_module;
  sys_call_table[129]= realsys_delete_module;
  sys_call_table[21] = realsys_mount;
  sys_call_table[22] = realsys_old_umount;
  sys_call_table[52] = realsys_umount;
  sys_call_table[37] = realsys_kill;
    
}
